dnl check_smlnj.m4
dnl
dnl COPYRIGHT (c) 2000 Bell Labs, Lucent Technologies.
dnl
dnl
dnl @synopsis AC_PROG_SMLNJ
dnl
dnl This macro figures out the location of SML/NJ and its major, minor,
dnl and patch version numbers.  The variables SMLNJ, SMLNJ_MAJOR_VERSION,
dnl SMLNJ_MINOR_VERSION, SMLNJ_PATCH_VERSION, SMLNJ_USING_NEW_CM, 
dnl and SMLNJ_CM_MAKE are set by this macro when it executes successfully.  
dnl This macro also does an AC_SUBST(SMLNJ) and AC_SUBST(SMLNJ_CM_MAKE). 
dnl You can override the version of SML/NJ used by defining the SMLNJ 
dnl variable in the  environment.
dnl 
dnl @version $Id: aclocal.m4,v 1.3 2001/07/17 00:18:16 danwang Exp $
dnl @author John Reppy <jhr@research.bell-labs.com>
dnl Modified by Daniel Wang
dnl
AC_DEFUN(AC_PROG_SMLNJ, [
dnl
dnl first we check for the existence of SML/NJ
dnl
  if test z$SMLNJ = z ; then
    AC_CHECK_PROGS(SMLNJ, sml-cm sml sml.bat, none)
  fi
  if test $SMLNJ = none; then
    AC_MSG_ERROR([sml not installed on PATH])
  else
dnl
dnl SML/NJ is installed, so determine the version numbers
dnl
   
   changequote(<<<,>>>)dnl
<<<$SMLNJ >/dev/null <<EOF
val _ = let
      val i2a = Int.toString
      val s = TextIO.openOut "conftest.sh"
      fun pr l = TextIO.output(s, concat l)
      in
        case #version_id Compiler.version
	 of [major] => (
	      pr["SMLNJ_MAJOR_VERSION=", i2a major, "\n"];
	      pr["SMLNJ_MINOR_VERSION=0\n"];
	      pr["SMLNJ_PATCH_VERSION=0\n"])
	  | [major, minor] => (
	      pr["SMLNJ_MAJOR_VERSION=", i2a major, "\n"];
	      pr["SMLNJ_MINOR_VERSION=", i2a minor, "\n"];
	      pr["SMLNJ_PATCH_VERSION=0\n"])
	  | (major::minor::patch::_) => (
	      pr["SMLNJ_MAJOR_VERSION=", i2a major, "\n"];
	      pr["SMLNJ_MINOR_VERSION=", i2a minor, "\n"];
	      pr["SMLNJ_PATCH_VERSION=", i2a patch, "\n"])
	  | _ => raise (Fail "unknown version")
	(* end case *);
      TextIO.closeOut s
      end;
EOF>>>
    changequote([, ])dnl
    . conftest.sh
    rm -fr "conftest.sh"
    AC_SUBST(SMLNJ)
  fi
dnl
dnl which version of CM are we using
dnl
  if test $SMLNJ_MAJOR_VERSION -eq 110 \
     -a $SMLNJ_MINOR_VERSION -eq 0 \
     -a $SMLNJ_PATCH_VERSION -ge 6
  then
    SMLNJ_USING_NEW_CM="FALSE"
  elif test $SMLNJ_MAJOR_VERSION -eq 110 \
    -a $SMLNJ_MINOR_VERSION -ge 32
  then
    SMLNJ_USING_NEW_CM="TRUE"
  else
    AC_MSG_ERROR([installation requires either version 110.0.6+ or 110.32+ of SML/NJ])
  fi
dnl
dnl if we are using the new version of CM, then we need to name
dnl the CM file explicitly.
dnl
  if test $SMLNJ_USING_NEW_CM = "TRUE" ; then
    SMLNJ_CM_MAKE="CM.make \"sources.cm\""
  else
    SMLNJ_CM_MAKE="CM.make ()"
  fi
  AC_SUBST(SMLNJ_CM_MAKE)
])dnl
dnl
AC_DEFUN(CHECK_SMLNJ_HEAP_SUFFIX, [
  AC_REQUIRE([AC_CANONICAL_HOST])
  case ${host_cpu}:${host_os} in
    alpha*:osf3*)	SMLNJ_ARCH=alpha32x;	SMLNJ_OPSYS=osf1;;
    alpha*:osf4*)	SMLNJ_ARCH=alpha32;	SMLNJ_OPSYS=dunix;;
    hppa*:hpux9*)	SMLNJ_ARCH=hppa;	SMLNJ_OPSYS=hpux9;;
    hppa*:hpux10*)	SMLNJ_ARCH=hppa;	SMLNJ_OPSYS=hpux;;
    i386:darwin*)	SMLNJ_ARCH=x86;		SMLNJ_OPSYS=darwin;;
    i386:freebsd*)	SMLNJ_ARCH=x86;		SMLNJ_OPSYS=freebsd;;
    i486:freebsd*)	SMLNJ_ARCH=x86;		SMLNJ_OPSYS=freebsd;;
    i586:freebsd*)	SMLNJ_ARCH=x86;		SMLNJ_OPSYS=freebsd;;
    i686:freebsd*)	SMLNJ_ARCH=x86;		SMLNJ_OPSYS=freebsd;;
    i386:linux*)	SMLNJ_ARCH=x86;		SMLNJ_OPSYS=linux;;
    i486:linux*)	SMLNJ_ARCH=x86;		SMLNJ_OPSYS=linux;;
    i586:linux*)	SMLNJ_ARCH=x86;		SMLNJ_OPSYS=linux;;
    i686:linux*)	SMLNJ_ARCH=x86;		SMLNJ_OPSYS=linux;;
    i386:netbsd*)	SMLNJ_ARCH=x86;		SMLNJ_OPSYS=netbsd;;
    i486:netbsd*)	SMLNJ_ARCH=x86;		SMLNJ_OPSYS=netbsd;;
    i586:netbsd*)	SMLNJ_ARCH=x86;		SMLNJ_OPSYS=netbsd;;
    i686:netbsd*)	SMLNJ_ARCH=x86;		SMLNJ_OPSYS=netbsd;;
    i386:solaris3)	SMLNJ_ARCH=x86;		SMLNJ_OPSYS=solaris;;
    mips:irix4*)	SMLNJ_ARCH=mipseb;	SMLNJ_OPSYS=irix4; HEAP_OPSYS=irix;;
    mips:irix5*)	SMLNJ_ARCH=mipseb;	SMLNJ_OPSYS=irix5; HEAP_OPSYS=irix;;
    mips:irix6*)	SMLNJ_ARCH=mipseb;	SMLNJ_OPSYS=irix6; HEAP_OPSYS=irix;;
    powerpc:aix*)	SMLNJ_ARCH=ppc;		SMLNJ_OPSYS=aix;;
    powerpc:darwin*)	SMLNJ_ARCH=ppc;		SMLNJ_OPSYS=darwin;;
    rs6000:aix*)	SMLNJ_ARCH=rs6000;	SMLNJ_OPSYS=aix;;
    sparc:solaris2*)	SMLNJ_ARCH=sparc;	SMLNJ_OPSYS=sunos;;
    sparc:solaris3*)	SMLNJ_ARCH=sparc;	SMLNJ_OPSYS=solaris;;
    *) $1 ;;
  esac
  if test z$SMLNJ_ARCH != z -a z$SMLNJ_OPSYS != z ; then
    if test z$HEAP_OPSYS = z ; then
      SMLNJ_HEAP_SUFFIX=${SMLNJ_ARCH}-${SMLNJ_OPSYS}
    else
      SMLNJ_HEAP_SUFFIX=${SMLNJ_ARCH}-${HEAP_OPSYS}
    fi
    AC_SUBST(SMLNJ_ARCH)
    AC_SUBST(SMLNJ_OPSYS)
    AC_SUBST(SMLNJ_HEAP_SUFFIX)
  fi
])dnl
